@@ -96,6 +96,7 @@ static gboolean spi_atk_bridge_signal_listener         (GSignalInvocationHint *s
 							gpointer               data);
 static gint     spi_atk_bridge_key_listener            (AtkKeyEventStruct     *event,
 							gpointer               data);
+static void     spi_atk_tidy_windows                   (void);
 
 /* For automatic libgnome init */
 extern void gnome_accessibility_module_init     (void);
@@ -395,6 +396,10 @@ spi_atk_bridge_exit_func (void)
     }
   this_app = NULL;
 
+  /*
+   * Check whether we still have windows which have not been deleted.
+   */
+  spi_atk_tidy_windows ();
   /*
    *  FIXME: this may be incorrect for apps that do their own bonobo
    *  shutdown, until we can explicitly shutdown to get the ordering
@@ -947,3 +952,35 @@ spi_atk_bridge_window_event_listener (GSignalInvocationHint *signal_hint,
 		       "window:%s", name);
   return TRUE;
 }
+
+static void
+spi_atk_tidy_windows (void)
+{
+  AtkObject *root;
+  gint n_children;
+  gint i;
+
+  root = atk_get_root ();
+  n_children = atk_object_get_n_accessible_children (root);
+  for (i = 0; i < n_children; i++)
+    {
+      AtkObject *child;
+      AtkStateSet *stateset;
+      CORBA_any any;
+      const gchar *name;
+     
+      child = atk_object_ref_accessible_child (root, i);
+      stateset = atk_object_ref_state_set (child);
+      
+      name = atk_object_get_name (child);
+      spi_init_any_string (&any, (char**) &name);
+      if (atk_state_set_contains_state (stateset, ATK_STATE_ACTIVE))
+        {
+          spi_atk_emit_eventv (G_OBJECT (child), 0, 0, &any, "window:deactivate");
+        }
+      g_free (stateset);
+
+      spi_atk_emit_eventv (G_OBJECT (child), 0, 0, &any, "window:destroy");
+      g_object_unref (child);
+    }
+}
