@@ -25,6 +25,7 @@
 
 #undef SPI_LISTENER_DEBUG
 #undef SPI_DEBUG
+#undef SPI_QUEUE_DEBUG
 
 #include <config.h>
 #ifdef SPI_DEBUG
@@ -41,6 +42,10 @@
 /* A pointer to our parent object class */
 static SpiListenerClass *spi_registry_parent_class;
 
+static GQuark _deactivate_quark = 0;
+static GQuark _activate_quark = 0;
+static GQuark _state_quark = 0;
+
 int _dbg = 0;
 
 typedef enum {
@@ -580,9 +585,6 @@ notify_listeners_cb (GList * const *list, gpointer user_data)
 {
   SpiListenerStruct *ls;
   NotifyContext     *ctx = user_data;
-#ifdef SPI_DEBUG
-  CORBA_string       s;
-#endif
 
   ls = (*list)->data;
 
@@ -590,23 +592,24 @@ notify_listeners_cb (GList * const *list, gpointer user_data)
   fprintf (stderr, "event quarks: %lx %lx %lx\n", ls->event_type_quark, ctx->etype.major, ctx->etype.minor);
   fprintf (stderr, "event name: %s\n", ctx->etype.event_name);
 #endif
-
   if ((ls->event_type_quark == ctx->etype.major) ||
       (ls->event_type_quark == ctx->etype.minor))
     {
 #ifdef SPI_DEBUG
+      CORBA_string s;
       fprintf (stderr, "notifying listener %d\n", 0);
 /* g_list_index (list, l->data)); */
       s = Accessibility_Accessible__get_name (ctx->source, ctx->ev);
       fprintf (stderr, "event source name %s\n", s);
       CORBA_free (s);
-#endif
-      
-      ctx->e_out.source = CORBA_Object_duplicate (ctx->source, ctx->ev);
       if (BONOBO_EX (ctx->ev))
         {
+	  CORBA_exception_free (ctx->ev);
           return SPI_RE_ENTRANT_CONTINUE;
 	}
+#endif
+      
+      ctx->e_out.source = ctx->source;
       
       if ((*list) && (*list)->data == ls)
         {
@@ -617,44 +620,214 @@ notify_listeners_cb (GList * const *list, gpointer user_data)
               DBG (1, g_warning ("Accessibility app error: exception during "
 		        "event notification: %s\n",
 		        CORBA_exception_id (ctx->ev)));
-	      if (ctx->ev->_major == CORBA_SYSTEM_EXCEPTION)
-		      CORBA_exception_init (ctx->ev);
-	      /* clear system exception on notify, it means listener is dead but
-	       * that's no concern of the event source :-) */
+	      CORBA_exception_free (ctx->ev);
+	      /* FIXME: check that this item is removed from the list
+	       * on system exception by a 'broken' listener */
 	    }
 	}
-      else /* dup re-entered */
-        {
-          CORBA_Object_release (ctx->e_out.source, ctx->ev);
-	}
     }  
 
   return SPI_RE_ENTRANT_CONTINUE;
 }
 
+static void
+registry_emit_event (SpiRegistry *registry, NotifyContext *ctx)
+{
+  GList       **list = get_listener_list (registry, ctx->etype.type_cat);
+
+  if (list && *list)
+    {
+    
+      spi_re_entrant_list_foreach (list, notify_listeners_cb, ctx);
+    }
+}
+
+static NotifyContext*
+registry_clone_notify_context (NotifyContext *ctx)
+{
+  NotifyContext *new_ctx = g_new0 (NotifyContext, 1);
+
+  new_ctx->ev = NULL;
+  new_ctx->source = bonobo_object_dup_ref (ctx->source, NULL); 
+  new_ctx->etype.event_name = CORBA_string_dup (ctx->etype.event_name);
+  new_ctx->etype.type_cat = ctx->etype.type_cat;
+  new_ctx->etype.major = ctx->etype.major;
+  new_ctx->etype.minor = ctx->etype.minor;
+  new_ctx->etype.detail = ctx->etype.detail;
+  new_ctx->e_out.type = CORBA_string_dup (ctx->e_out.type);
+  new_ctx->e_out.source = ctx->e_out.source;
+  new_ctx->e_out.detail1 = ctx->e_out.detail1;
+  new_ctx->e_out.detail2 = ctx->e_out.detail2;
+  CORBA_any__copy (&(new_ctx->e_out.any_data), &(ctx->e_out.any_data));
+  return new_ctx;
+}
+
+static void
+registry_flush_event_queue (SpiRegistry       *registry,
+			    gboolean      discard,
+			    CORBA_Environment *ev)
+{
+  NotifyContext *q_ctx;
+  while (!g_queue_is_empty (registry->deferred_event_queue)) {
+    q_ctx = g_queue_pop_tail (registry->deferred_event_queue);
+#ifdef SPI_QUEUE_DEBUG
+    fprintf (stderr, "%s! %s [n=%d] %p\n", (discard ? "discard" : "pop"), 
+	     q_ctx->etype.event_name, 
+	     (int) registry->deferred_event_queue->length, q_ctx);
+#endif
+    if (!discard)  {
+      q_ctx->ev = ev;
+      registry_emit_event (registry, q_ctx);
+    }
+    bonobo_object_release_unref (q_ctx->source, NULL);
+    CORBA_free ((void *)q_ctx->etype.event_name);
+    CORBA_free ((void *)q_ctx->e_out.type);
+    g_free (q_ctx);
+  }
+  registry->is_queueing = FALSE;
+}
+
+static gboolean
+registry_timeout_flush_queue (gpointer data)
+{
+  SpiRegistry *registry = data;
+  CORBA_Environment ev;
+#ifdef SPI_QUEUE_DEBUG
+  fprintf (stderr, "timeout! flushing queue...\n");
+#endif
+  CORBA_exception_init (&ev);
+  registry_flush_event_queue (registry, FALSE, &ev);
+  return FALSE;
+}
+
+static gboolean
+registry_discard_on_event (SpiRegistry *registry, NotifyContext *ctx)
+{
+  gboolean retval = FALSE;
+  NotifyContext *q_ctx = g_queue_peek_tail (registry->deferred_event_queue);
+  if ((q_ctx != NULL) &&
+      (ctx->etype.type_cat == ETYPE_WINDOW) && 
+      (ctx->etype.major == _activate_quark)) {
+    if (CORBA_Object_is_equivalent (ctx->source, q_ctx->source, NULL)) {
+      retval = TRUE;
+    }
+  }
+  return retval;
+}
+
+static gboolean
+registry_reset_on_event (SpiRegistry *registry, NotifyContext *ctx)
+{
+  return (ctx->etype.type_cat == ETYPE_WINDOW) ? TRUE : FALSE;
+}
+
+static void
+registry_start_queue (SpiRegistry *registry)
+{
+    g_timeout_add_full (G_PRIORITY_HIGH_IDLE, 
+			registry->exit_notify_timeout,
+			registry_timeout_flush_queue, registry, 
+			NULL);
+    registry->is_queueing = 1;
+}
+
+static gboolean
+registry_defer_on_event (SpiRegistry *registry, NotifyContext *ctx)
+{
+  gboolean defer = FALSE;
+  if ((ctx->etype.type_cat == ETYPE_WINDOW) && 
+      (ctx->etype.major == _deactivate_quark)) {
+    defer = TRUE;
+    registry_start_queue (registry);
+  }
+  /* defer all object:state-change events after a window:deactivate */
+  else if ((ctx->etype.type_cat == ETYPE_OBJECT) && 
+	   (ctx->etype.major == _state_quark)) {
+    defer = TRUE;
+  }
+  return defer;
+}
+
+static void
+registry_queue_event (SpiRegistry *registry, NotifyContext *ctx)
+{
+  NotifyContext *q_ctx = registry_clone_notify_context (ctx);
+#ifdef SPI_QUEUE_DEBUG
+    if (q_ctx->etype.type_cat != ETYPE_MOUSE)
+      fprintf (stderr, "push! %s %p\n", q_ctx->etype.event_name, q_ctx);
+#endif    
+  g_queue_push_head (registry->deferred_event_queue, q_ctx);
+}
+
+/**
+ * Dispose of event in one of several ways:
+ * 1) discard;
+ * 2) initiate queuing and push onto queue (below)
+ * 3) push on existing queue to either pop on timeout or on subsequent event
+ * 4) pass-through immediately
+ * 5) pass-through, discarding queued events
+ * 6) emit queued events and then pass through
+ **/
+static gboolean
+registry_filter_event (SpiRegistry *registry, NotifyContext *ctx,
+		       CORBA_Environment *ev)
+{
+  gboolean queue_is_empty = FALSE;
+  g_assert (ctx != NULL);
+
+  /* case #1 is not yet used */
+  if (registry_defer_on_event (registry, ctx)) { /* #2, #3 */
+    if (registry->is_queueing) {
+      registry_queue_event (registry, ctx);
+      return FALSE;
+    }
+    else { /* #4a */
+      return TRUE;
+    }
+  } 
+  else if (registry_reset_on_event (registry, ctx)) { /* #5, #6 */
+    gboolean discard = registry_discard_on_event (registry, ctx);
+#ifdef SPI_QUEUE_DEBUG
+    fprintf (stderr, "event %s caused reset, discard=%d\n",
+	     ctx->etype.event_name, (int) discard);
+#endif    
+    registry_flush_event_queue (registry, discard, ev);
+    return (discard ? FALSE : TRUE);
+  }
+  else { /* #4b */
+    return TRUE;
+  }
+}
+
 static void
 impl_registry_notify_event (PortableServer_Servant     servant,
                             const Accessibility_Event *e,
                             CORBA_Environment         *ev)
 {
   SpiRegistry  *registry;
-  GList       **list;
   NotifyContext ctx;
+  static int level = 0;
 
+  level++;
   registry = SPI_REGISTRY (bonobo_object_from_servant (servant));
 
   parse_event_type (&ctx.etype, e->type);
 
-  list = get_listener_list (registry, ctx.etype.type_cat);
+  ctx.ev = ev;
+  ctx.e_out = *e;
+  ctx.source = e->source;
 
-  if (list && *list)
-    {
-      ctx.ev = ev;
-      ctx.e_out = *e;
-      CORBA_any__copy (&ctx.e_out.any_data, &e->any_data);
-      ctx.source = e->source;
-      spi_re_entrant_list_foreach (list, notify_listeners_cb, &ctx);
-    }
+  if (registry_filter_event (registry, &ctx, ev)) {
+#ifdef SPI_QUEUE_DEBUG
+    if (ctx.etype.type_cat != ETYPE_MOUSE)
+{
+      fprintf (stderr, "emit! %s level: %d\n", ctx.etype.event_name, level);
+      fprintf (stderr, "emit! %p %p\n", ctx.e_out, ctx.e_out.type);
+}
+#endif    
+    registry_emit_event (registry, &ctx);
+  }
+  level--;
 }
 
 static void
@@ -678,15 +851,27 @@ spi_registry_class_init (SpiRegistryClass *klass)
   epv->getDesktopCount                  = impl_accessibility_registry_get_desktop_count;
   epv->getDesktop                       = impl_accessibility_registry_get_desktop;
   epv->getDesktopList                   = impl_accessibility_registry_get_desktop_list;
+  _deactivate_quark = g_quark_from_static_string ("deactivate");
+  _activate_quark = g_quark_from_static_string ("activate");
+  _state_quark = g_quark_from_static_string ("state-changed");
 }
 
 static void
 spi_registry_init (SpiRegistry *registry)
 {
   spi_registry_set_debug (g_getenv ("AT_SPI_DEBUG"));
+  /*
+   * TODO: fixme, this module makes the foolish assumptions that
+   * registryd uses the same display as the apps, and that the
+   * DISPLAY environment variable is set.
+   */
+  gdk_init (NULL, NULL);
+
   registry->object_listeners = NULL;
   registry->window_listeners = NULL;
   registry->toolkit_listeners = NULL;
+  registry->deferred_event_queue = g_queue_new ();
+  registry->exit_notify_timeout = 100;
   registry->desktop = spi_desktop_new ();
   /* Register callback notification for application addition and removal */
   g_signal_connect (G_OBJECT (registry->desktop),
@@ -705,7 +890,7 @@ spi_registry_init (SpiRegistry *registry)
 BONOBO_TYPE_FUNC_FULL (SpiRegistry,
 		       Accessibility_Registry,
 		       PARENT_TYPE,
-		       spi_registry);
+		       spi_registry)
 
 SpiRegistry *
 spi_registry_new (void)
